/*
 * moves_order.c --- General package to store moves order.
 *
 * Copyright (c) 2001, 2002, 2015 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of Natch.
 *
 * Natch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include        "common.h"

#include	"moves_order.h"

#include	"main.h"	/* for InitialePieces */
#include	"pcpj.h"	/* for NbCoupsBlancs/NoirsRestants */
#include        "pcpjtool.h"

#include	<assert.h>

/*--------------------------------------------------------------------------*/

typedef struct moves_list_t
{
    struct one_move_t*		move;
    struct moves_list_t*	next;
} moves_list_t;

typedef struct one_move_t
{
    piece_index_t	piece_index;
    unsigned int	move_index;
    move_type_t		type;

    moves_list_t*	successors_list;

    bool_t		print_value;
} one_move_t;

typedef struct moves_order_t
{
    moves_list_t*	moves_list;
} moves_order_t;

/*--------------------------------------------------------------------------*/

static bool_t		Print_Value = FALSE;

static moves_order_t	Moves_Order;

/*--------------------------------------------------------------------------*/
/* Specification of the allocation package for moves_list_t */

#define	MAX_MOVES_LISTS		200
static moves_list_t	Moves_Lists[ MAX_MOVES_LISTS ];
static moves_list_t*	Free_Moves_Lists;

static void		initialise_moves_list();
static moves_list_t*	new_moves_list();
static void		delete_moves_list(
    moves_list_t* moves_list );

/*--------------------------------------------------------------------------*/
/* Specification of the allocation package for one_move_t */

static void		initialise_list_of_moves();
static one_move_t*	new_one_move();

/*--------------------------------------------------------------------------*/

static void		print_one_move(
    FILE*			fd,
    const one_move_t*		move,
    const exploration_t*	exploration );

static void		print_successors_list(
    FILE*			fd,
    unsigned int		level,
    moves_list_t*		successors_list,
    const exploration_t*	exploration );

static one_move_t*	find_in_successors_list(
    moves_list_t*		successors_list,
    piece_index_t		piece_index,
    unsigned int		move_index,
    move_type_t			type );

static bool_t		is_equal(
    const one_move_t*		move_1,
    const one_move_t*		move_2 );

static void             remove_weak_constraints_internal(
    moves_list_t* moves_list,
    const exploration_t* exploration);

/*--------------------------------------------------------------------------*/

moves_order_t*
create_moves_order(exploration_t* exploration)
{
    unsigned int	i;
    moves_order_t*	moves_order = &Moves_Order;

    initialise_list_of_moves();
    initialise_moves_list();

    moves_order->moves_list = NULL;

    for ( i = INDEX_ROI_BLANC ; i <= INDEX_PION_NOIR_H ; i++ )
    {
        unsigned int	j;
        moves_list_t*	moves_list;
        piece_t*	piece 	      = &exploration->pieces[ i ];
        one_move_t*	previous_move = NULL;
        square_t	origin 	      = piece->caseInitiale;
        bool_t		hasPromoted   = FALSE;
        unsigned int    nb_spares =
            piece->camp == BLANC ? NbCoupsBlancsRestants : NbCoupsNoirsRestants;

        if (piece->destinations[0].distance == 0)
        {
            continue;		/* This piece never moves */
        }

        moves_list = new_moves_list();
        moves_list->move = NULL;
        moves_list->next = moves_order->moves_list;
        moves_order->moves_list = moves_list;

        for ( j = 0 ; j < piece->nbDestination ; j++ )
        {
            int deltaY = row(origin) - row(piece->destinations[j].destination);
            one_move_t*	move = new_one_move();

            piece->destinations[j].nb_predecessors_first = 0;
            piece->destinations[j].nb_predecessors_last = 0;
            piece->destinations[j].successors_first = NULL;
            piece->destinations[j].successors_last = NULL;

            move->piece_index 	  = i;
            move->move_index  	  = j;
            move->type        	  = e_First_Move;
            move->successors_list = NULL;
            move->print_value     = Print_Value;
            if ( previous_move == NULL )
            {
                moves_list->move = move;
            }
            else
            {
                previous_move->successors_list       = new_moves_list();
                previous_move->successors_list->move = move;
                previous_move->successors_list->next = NULL;
            }
            previous_move = move;

            if ( piece->destinations[j].distance == 1
                 && ( nb_spares == 0
                      || ( piece->destinations[j].destination ==
                           CASE_ARRIVEE_QUELCONQUE )
                      || ( piece->typePiece == ROI
                           && piece->castling != NO_CASTLING
                           && j == 0 )
                      || ( piece->typePiece == PION
                           && !hasPromoted
                           && ( deltaY == 16 || deltaY == -16 ) )
                      || ( piece->typePiece == CAVALIER
                           && nb_spares < 2 ) ) )
            {
                move->type = e_First_And_Last_Move;
            }
            else
            {
                move                  = new_one_move();
                move->piece_index     = i;
                move->move_index      = j;
                move->type            = e_Last_Move;
                move->successors_list = NULL;
                move->print_value     = Print_Value;

                previous_move->successors_list       = new_moves_list();
                previous_move->successors_list->move = move;
                previous_move->successors_list->next = NULL;
                previous_move                        = move;
            }
            origin = piece->destinations[j].destination;
            if ( origin == piece->casePromotion )
            {
                hasPromoted = TRUE;
            }
        }
    }
    return moves_order;
}


bool_t
add_move_order( moves_order_t*	moves_order,
                piece_index_t	before_piece_index,
                unsigned int	before_move_index,
                move_type_t	before_type,
                piece_index_t	after_piece_index,
                unsigned int	after_move_index,
                move_type_t	after_type )
{
    one_move_t*		before;
    one_move_t*		after;
    moves_list_t*	new_successor;
    moves_list_t*	moves_list;
    moves_list_t*	previous_in_list = NULL;

    if (before_piece_index == after_piece_index
        && before_move_index == after_move_index)
    {
        exit_with_error("add_move_order");
    }

    before = find_in_successors_list(
        moves_order->moves_list,
        before_piece_index,
        before_move_index,
        before_type );

    after = find_in_successors_list(
        moves_order->moves_list,
        after_piece_index,
        after_move_index,
        after_type );

    assert( before != NULL );
    assert( after  != NULL );

    if ( find_in_successors_list( after->successors_list,
                                  before_piece_index,
                                  before_move_index,
                                  before_type ) != NULL )
    {
        return FALSE;
    }

    if ( find_in_successors_list( before->successors_list,
                                  after_piece_index,
                                  after_move_index,
                                  after_type ) != NULL )
    {
        if ( MainPrintPositionLevel > 2 )
        {
            fprintf(MainFD, "=> already exists.");
        }
        return TRUE;
    }

    new_successor = new_moves_list();
    new_successor->move = after;
    new_successor->next = before->successors_list;
    before->successors_list = new_successor;

    /* Remove after from the top level list if it is there */
    moves_list = moves_order->moves_list;
    while ( moves_list != NULL )
    {
        if ( is_equal( after, moves_list->move ) )
        {
            if ( previous_in_list == NULL )
            {
                moves_order->moves_list = moves_list->next;
            }
            else
            {
                previous_in_list->next = moves_list->next;
            }
            delete_moves_list( moves_list );
            break;
        }
        previous_in_list = moves_list;
        moves_list = moves_list->next;
    }

    /* For each successors of before, if it is a successor of after, remove it
     * from the list the successors of before.
     */
    moves_list = before->successors_list;
    previous_in_list = NULL;
    while ( moves_list != NULL )
    {
        if ( find_in_successors_list( after->successors_list,
                                      moves_list->move->piece_index,
                                      moves_list->move->move_index,
                                      moves_list->move->type ) != NULL )
        {
            if ( previous_in_list == NULL )
            {
                before->successors_list = moves_list->next;
            }
            else
            {
                previous_in_list->next = moves_list->next;
            }
            new_successor = moves_list->next;
            delete_moves_list( moves_list );
            moves_list = new_successor;
        }
        else
        {
            previous_in_list = moves_list;
            moves_list = moves_list->next;
        }
    }
    return TRUE;
}

void
print_moves_order( FILE*		fd,
                   const moves_order_t*	moves_order,
                   const exploration_t*	exploration )
{
    assert( fd != NULL );
    assert( moves_order != NULL );

    Print_Value = ! Print_Value;

    print_successors_list( fd,
                           0,
                           moves_order->moves_list,
                           exploration );
}

/*--------------------------------------------------------------------------*/

static void
print_one_move( FILE*			fd,
                const one_move_t*	move,
                const exploration_t*	exploration )
{
    square_t	from;
    piece_t	piece = exploration->pieces[ move->piece_index ];
    static char	move_type[] = { ' ', 'F', 'L', 'B' };

    if ( move->move_index == 0 )
    {
        from = piece.caseInitiale;
    }
    else
    {
        from = piece.destinations[move->move_index - 1].destination;
    }

    fprintf( fd, "%c", InitialePieces[ piece.typePiece ] );
    if ( from != CASE_ARRIVEE_QUELCONQUE )
    {
        fprintf( fd, "%c%c", 'a' + column( from ), '1' + ( row( from ) >> 4 ) );
    }
    else
    {
        fprintf( fd, "??" );
    }
    if (piece.destinations[move->move_index].destination !=
        CASE_ARRIVEE_QUELCONQUE)
    {
        if ( piece.typePiece == ROI
             && move->move_index == 0
             && piece.castling != NO_CASTLING )
        {
            if ( piece.castling == KING_SIDE )
            {
                fprintf( fd, "-OO" );
            }
            else
            {
                fprintf( fd, "OOO" );
            }
        }
        else
        {
            fprintf( fd,
                     "-%c%c",
                     'a' +
                     column(piece.destinations[move->move_index].destination),
                     '1'
                     + (row(piece.destinations[move->move_index].destination) >>
                        4));
        }
    }
    else
    {
        fprintf( fd, "-??" );
    }
    fprintf( fd, "(%d,%d", move->piece_index, move->move_index );
    if ( move->type != e_First_And_Last_Move )
    {
        fprintf( fd, ",%c)", move_type[ move->type ] );
    }
    else
    {
        fprintf( fd, ")  " );
    }
}

static void
print_successors_list(
    FILE*			fd,
    unsigned int		level,
    moves_list_t*		successors_list,
    const exploration_t*	exploration )
{
    bool_t first = TRUE;
    while ( successors_list != NULL )
    {
        one_move_t* move = successors_list->move;

        if ( !first )
        {
            unsigned int	i;
            for ( i = 0 ; i < level ; i++ )
            {
                fprintf( fd, "                  " );
            }
        }
        else
        {
            if ( level > 0 )
            {
                fprintf( fd, " -> " );
            }
            first = FALSE;
        }

        /* If we can print in one time the First and Last part of the same move,
         * we do so.
         */
        if ( move->type == e_First_Move
             && move->successors_list != NULL
             && move->successors_list->next == NULL
             && move->piece_index == move->successors_list->move->piece_index )
        {
            move->type = e_First_And_Last_Move;
            print_one_move( fd, move, exploration );
            move->type = e_First_Move;

            move->print_value = Print_Value;
            move = move->successors_list->move;
        }
        else
        {
            print_one_move( fd, move, exploration );
        }

        if ( move->print_value != Print_Value )
        {
            if ( move->successors_list == NULL )
            {
                fputc( '\n', fd );
            }
            else
            {
                print_successors_list( fd,
                                       level + 1,
                                       move->successors_list,
                                       exploration );
            }
            move->print_value = Print_Value;
        }
        else
        {
            if ( move->successors_list != NULL )
            {
                fprintf( fd, " -> ..." );
            }
            fputc( '\n', fd );
        }

        successors_list = successors_list->next;
    }
}

/*--------------------------------------------------------------------------*/

static one_move_t*
find_in_successors_list(
    moves_list_t*	successors_list,
    piece_index_t	piece_index,
    unsigned int	move_index,
    move_type_t		type )
{
    while ( successors_list != NULL )
    {
        one_move_t* move = successors_list->move;

        if ( move->piece_index == piece_index
             && move->move_index == move_index
             && ( move->type & type ) != 0 )
        {
            return move;
        }

        move = find_in_successors_list( move->successors_list,
                                        piece_index,
                                        move_index,
                                        type );
        if ( move != NULL )
        {
            return move;
        }
        successors_list = successors_list->next;
    }
    return NULL;
}

/*--------------------------------------------------------------------------*/

static bool_t
is_equal( const one_move_t* move_1,
          const one_move_t* move_2 )
{
    assert( move_1 != NULL );
    assert( move_2 != NULL );

    return
        move_1->piece_index == move_2->piece_index
        && move_1->move_index == move_2->move_index
        && ( move_1->type & move_2->type ) != 0;
}

/*--------------------------------------------------------------------------*/
/* Allocation package for moves_list_t */

static void
initialise_moves_list()
{
    unsigned int i;
    for ( i = 0 ; i < MAX_MOVES_LISTS ; i++ )
    {
        Moves_Lists[ i ].next = &Moves_Lists[ i + 1 ];
    }
    Moves_Lists[ i - 1 ].next = NULL;
    Free_Moves_Lists = Moves_Lists;
}

static moves_list_t*
new_moves_list()
{
    moves_list_t* moves_list = Free_Moves_Lists;
    assert( Free_Moves_Lists != NULL );
    Free_Moves_Lists = Free_Moves_Lists->next;
    return moves_list;
}

static void
delete_moves_list( moves_list_t* moves_list )
{
    moves_list->next = Free_Moves_Lists;
    Free_Moves_Lists = moves_list;
}

/*--------------------------------------------------------------------------*/
/* Allocation package for one_move_t */

#define	MAX_MOVES	100
static one_move_t	List_Of_Moves[ MAX_MOVES ];
static unsigned int     Nb_Moves;

static void
initialise_list_of_moves()
{
    Nb_Moves = 0;
}

static one_move_t*
new_one_move()
{
    assert(Nb_Moves < MAX_MOVES);
    return &List_Of_Moves[Nb_Moves++];
}

void
shift_move_index(piece_index_t piece_index, unsigned int move_index, int delta)
{
    unsigned int i;

    for (i = 0; i < Nb_Moves; i++)
    {
        if (List_Of_Moves[i].piece_index == piece_index
            && List_Of_Moves[i].move_index >= move_index)
        {
            List_Of_Moves[i].move_index += delta;
        }
    }
}

/*--------------------------------------------------------------------------*/

void
remove_weak_constraints(
    moves_order_t* moves_order,
    const exploration_t* exploration)
{
    remove_weak_constraints_internal(moves_order->moves_list, exploration);
}

static void
remove_weak_constraints_internal(
    moves_list_t* moves_list,
    const exploration_t* exploration)
{
    while ( moves_list != NULL)
    {
        one_move_t* move = moves_list->move;
        moves_list_t* successors = move->successors_list;
        one_move_t* checked_moves[32];
        unsigned int nb_checked = 0;
        bool_t something_done = TRUE;

        while (something_done)
        {
            moves_list_t* successor = successors;
            moves_list_t* predecessor = NULL;
            bool_t already_done = TRUE;
            unsigned int i;

            something_done = FALSE;

            while (already_done &&successor != NULL)
            {
                already_done = FALSE;
                for (i = 0; i < nb_checked; i++)
                {
                    if (successor->move == checked_moves[i])
                    {
                        already_done = TRUE;
                        break;
                    }
                }
                if (already_done)
                {
                    predecessor = successor;
                    successor = successor->next;
                }
            }

            if (successor != NULL)
            {
                checked_moves[nb_checked] = successor->move;
                nb_checked++;
                something_done = TRUE;

                moves_list_t* other_successors = move->successors_list;
                while (other_successors != NULL)
                {
                    if (other_successors != successor)
                    {
                        one_move_t* a_move;
                        a_move = find_in_successors_list(
                                other_successors->move->successors_list,
                                successor->move->piece_index,
                                successor->move->move_index,
                                successor->move->type);
                        if (a_move != NULL)
                        {
                            if (predecessor == NULL)
                            {
                                move->successors_list = successor->next;
                                successors = successor->next;
                            }
                            else
                            {
                                predecessor->next = successor->next;
                            }
                            break;
                        }
                    }
                    other_successors = other_successors->next;
                }
            }
        }

        while (successors != NULL)
        {
            remove_weak_constraints_internal(successors, exploration);
            successors = successors->next;
        }

        moves_list = moves_list->next;
    }
}

/*--------------------------------------------------------------------------*/

static void
initialise_predecessors(exploration_t* exploration)
{
    piece_index_t index;

    for (index=INDEX_ROI_BLANC; index<=INDEX_PION_NOIR_H; index++)
    {
        piece_t* piece = &exploration->pieces[index];
        UCHAR i;

        for (i = 0; i < piece->nbDestination; i++)
        {
            destination_t* dest = &piece->destinations[i];
            dest->nb_predecessors_first = 0;
            dest->nb_predecessors_last = 0;
            dest->have_successors_decremented = FALSE;
            dest->is_successors_list_first_done = FALSE;
            dest->is_successors_list_last_done = FALSE;
            dest->successors_first = NULL;
            dest->successors_last = NULL;
        }
    }
}

static void
add_successors_list(
    exploration_t* exploration,
    destination_t* destination,
    moves_list_t* successors_list)
{
    while (successors_list != NULL)
    {
        one_move_t* move = successors_list->move;
        moves_list_t* next = successors_list->next;

        if (move->type == e_Last_Move)
        {
            successors_list->next = destination->successors_last;
            destination->successors_last = successors_list;

            exploration->pieces[move->piece_index]
                .destinations[move->move_index]
                .nb_predecessors_last++;
        }
        else
        {
            successors_list->next = destination->successors_first;
            destination->successors_first = successors_list;

            exploration->pieces[move->piece_index]
                .destinations[move->move_index]
                .nb_predecessors_first++;
        }

        successors_list = next;
    }
}

static moves_list_t*
eliminate_same_piece_successor(moves_list_t* successors, piece_index_t index)
{
    moves_list_t* previous = NULL;
    moves_list_t* result = successors;
    while (successors != NULL)
    {
        if (successors->move->piece_index == index)
        {
            if (previous == NULL)
            {
                result = successors->next;
            }
            else
            {
                previous->next = successors->next;
            }
        }
        previous = successors;
        successors = successors->next;
    }
    return result;
}

static void
update_predecessors_internal(
    moves_list_t* successors_list, exploration_t* exploration)
{
    while (successors_list != NULL)
    {
        one_move_t* move = successors_list->move;
        destination_t* destination = &exploration->pieces[move->piece_index]
            .destinations[move->move_index];

        if (!destination->is_successors_list_first_done
            && (move->type == e_First_Move
                || move->type == e_First_And_Last_Move))
        {
            destination->is_successors_list_first_done = TRUE;
            update_predecessors_internal(move->successors_list, exploration);
            add_successors_list(
                exploration,
                destination,
                eliminate_same_piece_successor(
                    move->successors_list,
                    move->piece_index));
        }
        else if (!destination->is_successors_list_last_done
                 && move->type == e_Last_Move)
        {
            destination->is_successors_list_last_done = TRUE;
            update_predecessors_internal(move->successors_list, exploration);
            add_successors_list(
                exploration,
                destination,
                eliminate_same_piece_successor(
                    move->successors_list,
                    move->piece_index));
        }

        successors_list = successors_list->next;
    }
}

static void
__attribute__ ((unused))
print_successors(const exploration_t* exploration)
{
    piece_index_t       index;

    for (index=INDEX_ROI_BLANC; index<=INDEX_PION_NOIR_H; index++)
    {
        piece_t piece = exploration->pieces[index];
        UCHAR i;

        for (i = 0; i < piece.nbDestination; i++)
        {
            destination_t dest = piece.destinations[i];
            moves_list_t* successor_first = dest.successors_first;
            moves_list_t* successor_last = dest.successors_last;

            if (successor_first != NULL || successor_last != NULL)
            {
                fprintf(MainFD, "(%d,%d) : ", index, i);
                while (successor_first != NULL)
                {
                    print_one_move(MainFD, successor_first->move, exploration);
                    fprintf(MainFD, "F");
                    successor_first = successor_first->next;
                }
                while (successor_last != NULL)
                {
                    print_one_move(MainFD, successor_last->move, exploration);
                    fprintf(MainFD, "L");
                    successor_last = successor_last->next;
                }
                fprintf(MainFD, "\n");
            }
        }
    }
}

static void
__attribute__ ((unused))
print_nb_predecessors(const exploration_t* exploration)
{
    piece_index_t index;

    for (index=INDEX_ROI_BLANC; index<=INDEX_PION_NOIR_H; index++)
    {
        piece_t piece = exploration->pieces[index];
        UCHAR i;

        for (i = 0; i < piece.nbDestination; i++)
        {
            destination_t dest = piece.destinations[i];
            if (dest.nb_predecessors_first > 0 || dest.nb_predecessors_last > 0)
            {
                fprintf(MainFD,
                        "%c%c%s(%d), nb_predecessors_first = %d, nb_predecessors_last = %d\n",
                        InitialePieces[piece.typePiece],
                        piece.camp == BLANC ? 'B' : 'N',
                        square_to_string(piece.caseInitiale),
                        i,
                        dest.nb_predecessors_first,
                        dest.nb_predecessors_last);
            }
        }
    }
}

void
create_predecessors(
    struct moves_order_t* moves_order, exploration_t* exploration)
{
    initialise_predecessors(exploration);
    update_predecessors_internal(moves_order->moves_list, exploration);
    /* print_successors(exploration); */
    /* print_nb_predecessors(exploration); */
}

void
update_predecessors_first_count(
    struct moves_list_t* successors, exploration_t* exploration, int value)
{
    while (successors != NULL)
    {
        exploration->pieces[successors->move->piece_index]
            .destinations[successors->move->move_index]
            .nb_predecessors_first += value;
        successors = successors->next;
    }
}

void
update_predecessors_last_count(
    struct moves_list_t* successors, exploration_t* exploration, int value)
{
    while (successors != NULL)
    {
        exploration->pieces[successors->move->piece_index]
            .destinations[successors->move->move_index]
            .nb_predecessors_last += value;
        successors = successors->next;
    }
}
